// This is a protobuf file. Pass it through the protobuf compiler to generate
// Python code:
//
// protoc --python_out=. metadata_analysis.proto

// Alternatively, run protobuf compiler from from Locan root directory:
// protoc --proto_path="." --python_out="."  "./locan/analysis/metadata_analysis.proto
//
// Update all *.proto files from Locan root directory:
// protoc --proto_path="." --proto_path="./locan/locan_io/locdata" --python_out="." "./locan/data/metadata.proto" "./locan/analysis/metadata_analysis.proto" "./locan/io/locdata/manifest.proto"
syntax = "proto2";

package locan.analysis;

import "google/protobuf/timestamp.proto";
// import "locan/data/metadata.proto";


message Analysis_routine {
    optional string name = 1;
    optional string parameter = 2;
}


// Next id: 8
message AMetadata {
    reserved 3, 4;
    reserved "creation_date", "modification_date";
    optional string identifier = 1;
    optional string comment = 2;
    // Localtime in "%Y-%m-%d %H:%M:%S %z"
    // deleted: optional string creation_date = 3;
    // Localtime in "%Y-%m-%d %H:%M:%S %z"
    // deleted: optional string modification_date = 4;
    optional Analysis_routine method = 5;
    // optional Metadata locdata_meta = 6;
    map<string, string> map = 7;
    optional google.protobuf.Timestamp creation_time = 8;
    optional google.protobuf.Timestamp modification_time = 9;
}